#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include "parse.h"
#include "eval.h"

/* #define debug_eval */

static ef_elem * ef_top = NULL;

static fun * fun_top = NULL;

static arg * arg_top = NULL;

static node * eval_depend_on_type(node * n);

void print_argument_stack(void) {
  arg * tmp;
  printf("argument stack:\n");
  for (tmp=arg_top;tmp!=NULL;tmp=tmp->next) {
    printf("'%s: %s\n",tmp->symbol,type_lookup(tmp->n->type));
  }
  printf("end of argument stack\n");
}

static void register_argument(char * symbol, node * n) {
  arg * a = xmalloc(sizeof(arg));
  a->symbol = symbol;
  a->n = n;
  a->next = arg_top;
  arg_top = a;
#ifdef debug_eval
  print_argument_stack();
#endif
  /*
  if (n->type == SYMBOL) {
    kill(getpid(),SIGSEGV);
  }
  */
#ifdef debug_eval
  printf("argument '%s registered (%d)\n",symbol,n->int_data);
  print_argument_stack();
#endif
}

node * get_argument(char * symbol) {
  arg * tmp;
  for (tmp=arg_top;tmp!=NULL;tmp=tmp->next) {
    if (strcasecmp(symbol,tmp->symbol)==0) {
      return tmp->n;
    }
  }
  return NULL;
}

static void unregister_argument(char * symbol) {
  arg * tmp = arg_top;
  if (strcasecmp(tmp->symbol,symbol)==0) {
    arg_top = tmp->next;
    free(tmp);
    return;
  }
  for (;tmp->next!=NULL;tmp=tmp->next) {
    if (strcasecmp(tmp->next->symbol,symbol)==0) {
      arg * tmp2 = tmp->next;
      tmp->next = tmp->next->next;
      free(tmp2);
      return;
    }
  }
  /* nothing found, nothing removed -- should we spit out a warning? */
}

node * eval_subtree(node * n) {
  if (n) {
    if (n->symbol_data) {
      node * (*eval_func)(char *, node **, int) = lookup_eval_func(n->symbol_data);
      if (eval_func==NULL) {
#ifdef debug_eval
        printf("eval_subtree: eval_func is NULL\n");
#endif
      }
      return eval_func(n->symbol_data,n->childs,n->child_count);
    }
    return n; /* without a symbol, we return the tree itself */
  }
  return NULL;
}

static node * eval_int(node * n) {
  if (n && n->type==INT) {
    return n;
  }
  /* maybe put automatic conversion here? */
  printf("eval_int: expected INT, got %s\n",type_lookup(n->type));
  exit(1);
  return NULL;
}

static node * eval_string(node * n) {
  if (n && n->type==STRING) {
    return n;
  }
  /* dito: put automatic conversion here? */
  printf("eval_string: expected STRING, got %s\n",type_lookup(n->type));
  exit(1);
  return NULL;
}

static node * eval_symbol(node * n) {
  if (n && n->type==SYMBOL) {
    arg * x;
    for (x=arg_top;x!=NULL;x=x->next) {
      if (strcasecmp(n->symbol_data,x->symbol)==0) {
#ifdef debug_eval
        printf("eval_symbol: %s: type of x->n is %s\n",n->symbol_data,type_lookup(x->n->type));
#endif
        return eval_depend_on_type(x->n);
      }
    }
#ifdef debug_eval
    printf("eval_symbol: return symbol itself\n");
#endif
    return n; /* when no symbol can be found in the list, return symbol itself */
  }
  return  NULL;
}

static node * eval_depend_on_type(node * n) {
  if (n) {
#ifdef debug_eval
    printf("eval_depend_on_type: type is %s\n",type_lookup(n->type));
#endif
    switch (n->type) {
    case SUB_TREE:
      return eval_subtree(n);
    case INT:
      return eval_int(n);
    case STRING:
      return eval_string(n);
    case SYMBOL:
      return eval_symbol(n);
    default:
      return NULL;
    }
  }
  return NULL;
}

void * lookup_eval_func(char * fn) {
  ef_elem * x;
#ifdef debug_eval
  printf("lookup_eval_func: trying to match %s\n",fn);
#endif
  if (!fn) {
    return NULL;
  }
  for (x=ef_top;x!=NULL;x=x->next) {
    if (strcasecmp(x->id,fn)==0) {
      return x->f;
    }
  }
  return NULL;
}

static node * eval_func_plus(char * symbol, node ** childs, int count) {
  int result = 0, i;
  node * n;
  node * x = new_node(0);
  for (i=0;i<count;i++) {
    node * o;
    n = childs[i];
    o = eval_depend_on_type(n);
    if (!o) {
      printf("plus: evaluated subtree is NULL\n");
      exit(1);
    }
    if (o->type == INT) {
      result += o->int_data;
    } else {
      printf("plus: expected INT, got %s\n",type_lookup(o->type));
      exit(1);
    }
  }
  x->type = INT;
  x->int_data = result;
  return x;
}

static node * eval_func_mul(char * symbol, node ** childs, int count) {
  int result = 1, i;
  node * n;
  node * x = new_node(0);
  for (i=0;i<count;i++) {
    node * o;
    n = childs[i];
    o = eval_depend_on_type(n);
    if (!o) {
      printf("mul: evaluated subtree is NULL\n");
      exit(1);
    }
    if (o->type == INT) {
      result *= o->int_data;
    } else {
      printf("mul: expected INT, got %s\n",type_lookup(o->type));
      exit(1);
    }
  }
  x->type = INT;
  x->int_data = result;
  return x;
}

static node * eval_func_minus(char * symbol, node ** childs, int count) {
  node * n1, * n2;
  node * result;
  if (count!=2) {
    printf("minus: expected 2 arguments, got %d\n",count);
    exit(1);
  }
#ifdef debug_eval
  printf("argument 1: type is %s\n",type_lookup(childs[0]->type));
  printf("argument 2: type is %s\n",type_lookup(childs[1]->type));
#endif
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type!=INT) {
    printf("minus: argument 1 is %s, expected INT\n",type_lookup(n1->type));
    exit(1);
  }
  if (n2->type!=INT) {
    printf("minus: argument 2 is %s, expected INT\n",type_lookup(n2->type));
    exit(1);
  }
  result = new_node(0);
  result->type = INT;
  result->int_data = n1->int_data - n2->int_data;
  return result;
}


static node * eval_func_div(char * symbol, node ** childs, int count) {
  node * n1, * n2;
  node * result;
  if (count!=2) {
    printf("div: expected 2 arguments, got %d\n",count);
    exit(1);
  }
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type!=INT) {
    printf("div: argument 1 is %s, expected INT\n",type_lookup(n1->type));
    exit(1);
  }
  if (n2->type!=INT) {
    printf("div: argument 2 is %s, expected INT\n",type_lookup(n2->type));
    exit(1);
  }
  result = new_node(0);
  result->type = INT;
  result->int_data = n1->int_data / n2->int_data;
  return result;
}

static node * eval_func_userdefined(char * symbol, node ** childs, int count) {
  fun * tmp;
  node * fhead;
  node ** fbody;
  node * result = NULL;
  int bodycount;
  int i;
  for (tmp=fun_top;tmp!=NULL;tmp=tmp->next) {
    if (strcasecmp(symbol,tmp->name)==0) {
      break;
    }
  }
  if (!tmp) {
    printf("unable to find a function definition for %s\n",symbol);
    exit(1);
  }
  fhead = tmp->head;
  if (count!=fhead->child_count) {
    printf("argument count mismatch: expected %d arguments for %s, got %d arguments\n",fhead->child_count,symbol,count);
    exit(1);
  }
  for (i=0;i<fhead->child_count;i++) {
    node * x = NULL;
    if (fhead->childs[i]->type != SYMBOL) {
      printf("argument declaration %d of %s is not a symbol\n",i,symbol);
      exit(1);
    }
    if (childs[i]->type==SYMBOL) {
      x = get_argument(childs[i]->symbol_data);
    } else {
      x = childs[i];
    }
    if (x->type==SUB_TREE) {
      x = eval_subtree(x);
    }
    register_argument(fhead->childs[i]->symbol_data,x); /* XXX */
  }
  fbody = tmp->body;
  bodycount = tmp->bodycount;
  for (i=0;i<bodycount;i++) {
    result = eval_depend_on_type(fbody[i]); /* memory leak!!! */
  }
  for (i=0;i<fhead->child_count;i++) {
    unregister_argument(fhead->childs[i]->symbol_data);
  }
  return result;
}


static void register_function(char * fname, node * fhead, node ** fbody, int bodycount) {
  fun * function = xmalloc(sizeof(fun));
  function->name = strdup(fname);
  function->head = fhead;
  function->body = fbody;
  function->bodycount = bodycount;
  function->next = fun_top;
  fun_top = function;
  /* register user-defined function;
   * this is handled by eval_func_userdefined
   */
  add_eval_func(function->name,eval_func_userdefined);
}

static node * eval_func_defun(char * symbol, node ** childs, int count) {
  node * fhead;
  if (count<2) {
    printf("defun: function body missing\n");
    exit(1);
  }
  fhead = childs[0];
  if (fhead->symbol_data) {
    register_function(fhead->symbol_data,fhead,childs+1,count-1);
  } else {
    printf("defun: empty function head\n");
    exit(1);
  }
  return NULL;
}

static node * eval_func_if(char * symbol, node ** childs, int count) {
  node * cond = NULL, * ifyes = NULL, * ifno = NULL;
  if (count<2) {
    printf("if: too few arguments\n");
    exit(1);
  }
  cond = childs[0];
  ifyes = childs[1];
  if (count==3) {
    ifno = childs[2];
  }
  cond = eval_depend_on_type(cond);
  if (!cond) {
    printf("if: condition evaluated to NULL\n");
    exit(1);
  }
  if (cond->type!=INT) {
    printf("if: condition must be of type INT\n");
    exit(1);
  }
  if (cond->int_data) {
    return eval_depend_on_type(ifyes);
  } else if (ifno) {
    return eval_depend_on_type(ifno);
  }
  return NULL; /* shouldn't be */
}

node * eval_func_eq(char * symbol, node ** childs, int count) {
  node * n1, * n2, * retnode;
  int retval = 0;
  if (count!=2) {
    printf("eq: need two arguments\n");
    exit(1);
  }
  n1 = eval_depend_on_type(childs[0]);
  n2 = eval_depend_on_type(childs[1]);
  if (n1->type != n2->type) {
    printf("eq: type mismatch. can't compare %s and %s\n",type_lookup(n1->type),type_lookup(n2->type));
    exit(1);
  }
  switch (n1->type) {
  case SUB_TREE:
    /* unimplemented - how should it be done? */
    retval = 0; break;
  case INT:
    if (n1->int_data == n2->int_data) {
      retval = 1;
    } else {
      retval = 0;
    }
    break;
  case STRING:
    if (strcmp(n1->str_data,n2->str_data)==0) {
      retval = 1;
    } else {
      retval = 0;
    }
    break;
  case SYMBOL:
    if (strcasecmp(n1->symbol_data,n2->symbol_data)==0) {
      retval = 1;
    } else {
      retval = 0;
    }
    break;
  default:
    printf("unknown type %d\n",n1->type);
  }
  retnode = new_node(0);
  retnode->int_data = retval;
  retnode->type = INT;
#ifdef debug_eval
  printf("eq: return value is %d\n",retval);
#endif
  return retnode;
}

void init_eval_system(void) {
  add_eval_func("plus",eval_func_plus);
  add_eval_func("minus",eval_func_minus);
  add_eval_func("mul",eval_func_mul);
  add_eval_func("div",eval_func_div);
  add_eval_func("defun",eval_func_defun);
  add_eval_func("if",eval_func_if);
  add_eval_func("eq",eval_func_eq);
}

void add_eval_func(char * id, void * fptr) {
  ef_elem * x = xmalloc(sizeof(ef_elem));
  x->f = fptr;
  x->id = strdup(id);
  x->next = ef_top;
  ef_top = x;
}

